/* Copyright (c) 2005-2010 Ondrej Donek.
 * See the file LICENSE.txt for licensing information. */

// require global.js
// require bookmarks.js
// require templates.js
(function() {
	
	/** 
	 * @private
	 * @type String
	 */
	const kXUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";
	
	/** @type Components.interfaces.nsIPrefBranch */
	const kATOM_SRV = Components.classes["@mozilla.org/atom-service;1"].
			getService(Components.interfaces.nsIAtomService);
	
	/** @type Components.interfaces.nsICategoryManager */
	const kCM_SRV = Components.classes["@mozilla.org/categorymanager;1"].
			getService(Components.interfaces.nsICategoryManager);
	
	/** @type Components.interfaces.nsIPrefBranch */
	const kPREF_SRV = Components.classes["@mozilla.org/preferences-service;1"].
			getService(Components.interfaces.nsIPrefService).
			getBranch("extensions.xpcomviewer.");
	
	
	/**
	 * Localized titles for tree tabs (Filled from string bundles 
	 * in xpcomviewer.onLoad()).
	 * @private
	 * @type Object
	 */
	var _viewLabels = {
		classes: "Classes",
		ifaces:  "Interfaces", 
		results: "Results", 
		compman: "Manager entries"
	};
	
	/**
	 * Holds type of currently displayed items ['classes','ifaces','results','compman']
	 * @private
	 * @type String
	 */
	var _treeType = "";
	
	/**
	 * Holds currently visible tree items.
	 * @private
	 * @type Object
	 */
	var _treeItems = {
		classes: [],
		ifaces: [], 
		results: [], 
		compman: []
	};
	
	/**
	 * Back-up for original tree items (with no filter on it).
	 * @private
	 * @type Object
	 */
	var _originalTreeItems = {
		classes: [],
		ifaces: [], 
		results: [], 
		compman: []
	};
	
	/**
	 * Reference to tree element.
	 * @private
	 * @type DOMElement
	 */
	var _tree = null;
	
	/**
	 * Holds reference to filter textbox.
	 * @private
	 * @type DOMElement
	 */
	var _filter = null;
	
	/**
	 * Holds currnetly used filter string.
	 * @private
	 * @type String
	 */
	var _filterString = null;
	
	
	/**
	 * Prototype object for single tree item (single tree row).
	 *
	 * @param aName {string} XPCOM item name.
	 * @param aValue {string} XPCOM item value.
	 * @param aType {string} Type of XPCOM item (class, interface etc.)
	 * @param aLevel {integer} Level of item in tree.
	 * @param aIsContainer {boolean} Is TRUE if tree item has any sub-rows.
	 * @param aIsEmpty {boolean} If item is a container is empty or not?
	 */
	function XV_TreeitemPrototype(aName, aValue, aType, aLevel, aIsContainer, aIsEmpty)
	{
		this.name = aName;
		this.val = aValue;
		this.type = aType;
		this.level = aLevel;
		this.isCont = aIsContainer;
		this.isEmpty = aIsEmpty;
		this.isOpen = false;
		//this._hiddenRows = new Array();
	}; // End of XV_TreeitemPrototype(aName, aValue, aType, aLevel, aIsContainer)
	
	
	/**
	 * Prototype object for XPCOMViewer treeviews.
	 * 
	 * @returns {void}
	 */
	function XV_TreeviewPrototype() 
	{
		this.treebox = null;
		this.selection = null;
		this.sorted = false;
		this.sortDir = 0;
		this.sortedCol = "";
	};
	XV_TreeviewPrototype.prototype = 
	{
		get rowCount() { 
			return _treeItems[_treeType].length; 
		},
		
		getCellText: function(aIndex, aCol) {
			if(!(aIndex in _treeItems[_treeType])) return "";
			
			switch(aCol.id) {
				case "bookmarkCol": return ""; break;
				case "nameCol": return _treeItems[_treeType][aIndex].name; break;
				case "valueCol": return _treeItems[_treeType][aIndex].val; break;
			}
			
			return "";
		},
		
		getCellValue: function(aIndex, aCol) {
			if(!(aIndex in _treeItems[_treeType])) return null;
			
			return _treeItems[_treeType][aIndex].type;
		},
		
		getLevel: function(aIndex) {
			if(!(aIndex in _treeItems[_treeType]) || _treeType == "results") return -1;
			
			return _treeItems[_treeType][aIndex].level;
		},
		
		isContainer: function(aIndex) {
			if(!(aIndex in _treeItems[_treeType]) || _treeType == "results") return false;
			
			return _treeItems[_treeType][aIndex].isCont;
		},
		
		isContainerOpen: function(aIndex) {
			if(!(aIndex in _treeItems[_treeType]) || _treeType == "results") return false;
			if(_treeItems[_treeType][aIndex].isCont != true) return false;
			
			return _treeItems[_treeType][aIndex].isOpen;
		},
		
		isContainerEmpty: function(aIndex) {
			if(!(aIndex in _treeItems[_treeType]) || _treeType == "results") return false;
			if(_treeItems[_treeType][aIndex].isCont != true) return false;
			
			return _treeItems[_treeType][aIndex].isEmpty;
		},
		
		setTree: function(aOut) {
			this.treebox = aOut;
		},
		
		getParentIndex: function(aIndex) {
			if(!(aIndex in _treeItems[_treeType]) || _treeType == "results") return -1;
			
			let current = this.getLevel(aIndex); // current level
			
			for(let i = (aIndex - 1); i >= 0; i--) 
				if(this.getLevel(i) < current) 
					return i;
			
			return -1;
		},
		
		hasNextSibling: function(aIndex, aAfterIndex) {
			if(!(aIndex in _treeItems[_treeType])) return false;
			
			let count = this.rowCount; // total count of tree items
			let current = this.getLevel(aIndex); // current level
			
			if(_treeType == "results") {
				return ((aIndex + 1) < count);
			}
			
			for(let i = (aIndex + 1); i < count; i++) {
				let next = this.getLevel(i); // next level

				if(next == current) return true;
				else if(next < current) return false;
			}

			return false;
		}, // end hasNextSibling(aIndex, aAfterIndex)
		
		toggleOpenState: function(aIndex) {
			if(!(aIndex in _treeItems[_treeType]) || _treeType == "results") return;
			if(this.isContainer(aIndex) != true) return;
			
			let name = _treeItems[_treeType][aIndex].name;
			
			if(this.isContainerOpen(aIndex)) { // we are closing opened container
				let delCount = 0;
				let current = this.getLevel(aIndex);

				for(let i = (aIndex + 1); i < this.rowCount; i++) {
					if(this.getLevel(i) > current) delCount++;
					else break;
				}

				if(delCount > 0) {
					_treeItems[_treeType].splice(aIndex + 1, delCount);
					this.treebox.rowCountChanged(aIndex + 1, -delCount);
				}

				_treeItems[_treeType][aIndex].isOpen = false;
				return;
			} 
			
			// we are opening closed container
			if(_treeType == "classes") { // Tree "Components.classes"
				if(_treeItems[_treeType][aIndex].type == "class") { // Is selected class
					let cl = null;
					try { cl = Components.classes[name].createInstance(); } catch(e) { return; }
					
					for(let ifc in Components.interfaces) {
						if(ifc != "QueryInterface" && 
						   ifc != "canCreateWrapper" && 
						   ifc != "canCallMethod" && 
						   ifc != "canGetProperty" && 
						   ifc != "canSetProperty") {
							try {
								if(cl instanceof Components.interfaces[ifc]) { 
									let item = new XV_TreeitemPrototype(
											ifc, 
											ifc, 
											"iface_init", 
											1, 
											(ifc != "nsISupports" && ifc != "IDispatch"), 
											!(ifc != "nsISupports" && ifc != "IDispatch")
									);
									
									_treeItems[_treeType].splice(aIndex + 1, -1, item);
									this.treebox.rowCountChanged(aIndex + 1, 1);
								}
							} catch(e) {
								Components.utils.reportError(e);
							}
						}
					}
					_treeItems[_treeType][aIndex].isOpen = true;
				} else if(_treeItems[_treeType][aIndex].type == "iface_init" && name != "nsISupports")  { // Is selected interface
					let parentIndex = this.getParentIndex(aIndex);
					let clName = _treeItems[_treeType][parentIndex].name;
					let inst = null;
					
					try {
						inst = Components.classes[clName].createInstance(Components.interfaces[name]);
					} catch(e) {
						try { inst = Components.classes[clName].getService(Components.interfaces[name]); } catch(e) { return; }
					}

					if(inst) {
						try {
							for(var _prop in inst) {
								let prop = inst[_prop];
								let type = ((new String(prop)).indexOf("function") != -1) ? "iface_method" : "iface_prop";
								let val = (type == "iface_method") ? "" : prop;
								let item = new XV_TreeitemPrototype(_prop, val, type, 2, false, false);

								_treeItems[_treeType].splice(aIndex + 1, -1, item);
								this.treebox.rowCountChanged(aIndex + 1, 1);
							}
							_treeItems[_treeType][aIndex].isOpen = true;
						} catch(e) {
							Components.utils.reportError(e);
						}
					}
				}
			}
			else if(_treeType == "ifaces") { // Tree "Components.interfaces"
				_treeItems[_treeType][aIndex].isOpen = true;
				let ifc = Components.interfaces[name];

				for(let name in ifc) {
					let cnst = Components.interfaces[ifc][name];
					let item = new XV_TreeitemPrototype(name, cnst, "iface_const", 1, false, false);
					_treeItems[_treeType].splice(aIndex + 1, -1, item);
					this.treebox.rowCountChanged(aIndex + 1, 1);
				}
			} 
			else if(_treeType == "compman") { // Tree Component manager
				if(_treeItems[_treeType][aIndex].type == "compman_category") {
					_treeItems[_treeType][aIndex].isOpen = true;
					let entries = kCM_SRV.enumerateCategory(name);
					while(entries.hasMoreElements()) {
						let _entry = entries.getNext();
						let entry = _entry.QueryInterface(Components.interfaces.nsISupportsCString);
						let item = new XV_TreeitemPrototype(entry, "", "compman_entry", 1, false, false);
						_treeItems[_treeType].splice(aIndex + 1, -1, item);
						this.treebox.rowCountChanged(aIndex + 1, 1);
					}
				}
			}
		}, // end toggleOpenState(aIndex)
		
		cycleHeader: function(aCol) {
			let tree = document.getElementById("tree");
			let index = tree.currentIndex;

			// We have to ensure that all containers are closed
			for(let i=0; i<this.rowCount; i++) {
				if(_treeItems[_treeType][i].isCont && _treeItems[_treeType][i].isOpen) {
					this.toggleOpenState(i);
				}
			}

			if(aCol.id == this.sortedCol) {
				this.sortDir = (this.sortDir == 0)
						? 1
						: -this.sortDir;
			}
			
			if(aCol.id == this.sortedCol &&
				 this.fastIndex == this.rowCount) {
				_treeItems[_treeType].reverse();
			} else {
				if(this.sortedCol !== "") {
					let old = document.getElementById(this.sortedCol);
					if(old) {
						old.setAttribute("sortDirection", "");
					}
				}

				this.sortedCol = aCol.id;
				_treeItems[_treeType].sort(function(x, y) {
					if(x > y) return 1;
					else if(x < y) return -1;
					else return 0;
				});
			}

			let dir = (this.sortDir == 1) ? "ascending" : "descending";

			aCol.element.setAttribute("sortDirection", dir);
			this.treebox.invalidate();

			if(index >= 0) {
				this.selection.select(index);
				this.treebox.ensureRowIsVisible(index);
			}

			this.fastIndex = this.rowCount;
			this.sorted = true;
		}, // end cycleHeader(aCol)
		
		getCellProperties: function(aIndex, aCol, aProp) {
			if(!(aIndex in _treeItems[_treeType])) {
				aProp.AppendElement(kATOM_SRV.getAtom("noIcon"));
			}
			if(_treeType == "results") {
				aProp.AppendElement(kATOM_SRV.getAtom(_treeItems[_treeType][aIndex].type));
			}
			
			let name = _treeItems[_treeType][aIndex].name;
					
			if(aCol.id == "bookmarkCol") {
				// TODO Tohle zjistovani bookmarks je docela neefektivni - zkousel jsem 
				//      pridat jednoduchou cache, ale to se zase pak musel treeview 
				//      porad obnovovat...
				//if(typeof(_treeItems[_treeType][aIndex].bookmark) == "undefined") {
				//	_treeItems[_treeType][aIndex].bookmark = (XV_Utils.getBookmark(name) == null);
				//}
				if(xpcomviewer.bookmarks.hasBookmark(name))
					aProp.AppendElement(kATOM_SRV.getAtom("hasBookmark"));
				else 
					aProp.AppendElement(kATOM_SRV.getAtom("noBookmark"));
				
			} else if(aCol.id == "nameCol")  {
				if(_treeItems[_treeType][aIndex].type == "interface") {
					if(name == "nsISupports" || name == "nsIClassInfo" || name == "IDispatch") {
						aProp.AppendElement(kATOM_SRV.getAtom("constIface"));
						return;
					}
				} else if(_treeItems[_treeType][aIndex].type == "method") {
					if(name == "QueryInterface") {
						aProp.AppendElement(kATOM_SRV.getAtom("constMethod"));
						return;
					}
				}
				
				aProp.AppendElement(kATOM_SRV.getAtom(_treeItems[_treeType][aIndex].type));
			} else {
				aProp.AppendElement(kATOM_SRV.getAtom("noIcon"));
			}
		}, // end getCellProperties(aIndex, aCol, aProp)
		
		getRowProperties: function(aIndex, aProp) { return; },
		getColumnProperties: function(aCol, aProp) { return; },
		canDrop: function(aIndex, aOrientation) { return true; },
		drop: function(aIndex, aOrientation) {},
		getImageSrc: function(aIndex, aCol) { return ""; },
		
		isSorted: function() { 
			return this.sorted; 
		}, // end isSorted()
		
		selectionChanged: function() {},
		cycleCell: function(aIndex, aCol) {},
		isEditable: function(aIndex, aCol) { return false; },
		setCellText: function(aIndex, aCol, aValue) {},
		setCellValue: function(aIndex, aCol, aValue) {},
		performAction: function(aAction) {},
		performActionOnRow: function(aAction, aIndex) {},
		preformActionOnCell: function(aAction, aIndex, aCol) {},
		isSeparator: function(aIndex) { return false; }
	}; // End of XV_TreeviewPrototype
	
	
	/**
	 * On load event handler
	 *
	 * @param aEvent {Components.interfaces.nsIDOMEvent}
	 * @returns {void}
	 */
	this.onLoad = function(aEvent) 
	{
		// Initialize string bundles and load all needed localized strings
		let strBundle = document.getElementById("stringbundle");
		_viewLabels = {
				classes: strBundle.getString("classesLabel"),
				ifaces : strBundle.getString("ifacesLabel"),
				results: strBundle.getString("resultsLabel"),
				compman: strBundle.getString("compmanLabel")
		};
		
		xpcomviewer.bookmarks.prepareDb();
		xpcomviewer.templates.prepareDb();
		
		// Update view if neccessary
		if(kPREF_SRV.getBoolPref("useLastView")) {
			// User wants to restore view from the latest session
			this.showView(kPREF_SRV.getCharPref("lastViewName"));
		}/* 
		else if(kPREF_SRV.getBoolPref("useCustomView")) {
			// User wants to show the custom view
			this.showView(kPREF_SRV.getCharPref("customViewName"));
		} */
		else {
			this.showView("classes");
		}
	}; // end onLoad(aEvent)
	
	
	/**
	 * On unload event handler
	 *
	 * @param aEvent {Components.interfaces.nsIDOMEvent}
	 * @returns {Boolean}
	 */
	this.onUnload = function(aEvent) 
	{
		if(kPREF_SRV.getBoolPref("useLastView")) {
			kPREF_SRV.setCharPref("lastViewName", _treeType);
		}
		
		return true;
	}; // end onUnLoad(aEvent)
	
	
	/**
	 * Returns name of selected treeitem (if no selected returns empty string).
	 * 
	 * @returns {String}
	 */
	this.getSelectedTreeItemName = function()
	{
		if(_tree == null) {
			_tree = document.getElementById("tree");
		}
		
		if(_tree.currentIndex == -1) return "";
		
		return _tree.view.getCellText(
				_tree.currentIndex, 
				_tree.columns.getPrimaryColumn()
		);
	}; // end getSelectedTreeItemName()
	
	
	/**
	 * Returns selected treeitem as simple object (if no treeitem is selected 
	 * returns NULL).
	 * 
	 * @returns {Object}
	 */
	this.getSelectedTreeItem = function()
	{
		if(_tree == null) {
			_tree = document.getElementById("tree");
		}
		
		if(_tree.currentIndex == -1) return null;
		
		let col = _tree.columns.getPrimaryColumn();
		let itemIdx = _tree.currentIndex;
		let itemName = _tree.view.getCellText(itemIdx, col);
		let itemType =  _tree.view.getCellValue(itemIdx, col);
		
		return { idx: itemIdx, name: itemName, type: itemType };
	}; // end getSelectedTreeItem()
	
	
	/**
	 * Returns parent of selected treeitem as simple object (if no treeitem 
	 * is selected or selected item has no parent returns NULL).
	 * 
	 * @returns {Object}
	 */
	this.getSelectedTreeItemParent = function()
	{
		let item = this.getSelectedTreeItem();
		if(item == null) return null;
		
		if(_tree == null) {
			_tree = document.getElementById("tree");
		}
		
		let col = _tree.columns.getPrimaryColumn();
		let parentIdx = _tree.view.getParentIndex(item.idx);
		let parentName = _tree.view.getCellText(parentIdx, col);
		let parentType =  _tree.view.getCellValue(parentIdx, col);
		
		return { idx: parentIdx, name: parentName, type: parentType };
	}; // end getSelectedTreeItemParent()
	
	
	/**
	 * Show specified view ['classes', 'ifaces', 'results', 'compman'].
	 * 
	 * @param aView {String}
	 * @returns {void}
	 */
	this.showView = function(aView) 
	{
		if(_treeType == aView) return;
		
		if(_tree == null) {
			_tree = document.getElementById("tree");
		}
		
		this.loadTreeitems(aView);
		
		let prevType = _treeType;
		_treeType = aView;
		
		try {
			//_tree.treeBoxObject.view.selection.clearSelection();
			_tree.view.selection.clearSelection();
			_tree.view = new XV_TreeviewPrototype();
		} catch(e) {
			Components.utils.reportError(e);
			_treeType = prevType;
			return;
		}
		
		// Update toolbar buttons
		let btn = document.getElementById(aView + "-tbtn");
		if(btn) btn.checked = true;
		
		_tree.focus();
	}; // end showView(aView)
	
	
	/**
	 * Loads treeitems for the selected view.
	 * 
	 * @param aView {String}
	 * @returns {array}
	 */
	this.loadTreeitems = function(aView)
	{
		if(_treeItems[aView].length > 0) {
			return;
		}
		
		// Load Components.classes
		if(aView == "classes") {
			for(let cl in Components.classes) {
				if(typeof(Components.classes[cl].number) == "undefined") continue;
				
				_treeItems.classes.push(new XV_TreeitemPrototype(
						Components.classes[cl].name, 
						Components.classes[cl].number, 
						"class", 
						0, 
						true, 
						false
				));
			}
			
			if(_originalTreeItems.classes.length == 0) {
				_originalTreeItems.classes = _treeItems.classes;
			}
		}
		
		// Load Components.interfaces
		if(aView == "ifaces") {
			for(let ifc in Components.interfaces) {
				if(typeof(Components.interfaces[ifc].name) == "undefined") continue;

				let num = (typeof(Components.interfaces[ifc].number) == "undefined") 
						? "" 
						: Components.interfaces[ifc].number;
				
				let isCont = false;
				for(let ifcItem in Components.interfaces[ifc]) isCont = true;
				
				_treeItems.ifaces.push(new XV_TreeitemPrototype(
						Components.interfaces[ifc].name, 
						num, 
						"iface", 
						0, 
						isCont, 
						!isCont
				));
			}
			
			if(_originalTreeItems.ifaces.length == 0) {
				_originalTreeItems.ifaces = _treeItems.ifaces;
			}
		}
		
		// Load Components.results
		if(aView == "results") {
			for(let res in Components.results) {
				_treeItems.results.push(new XV_TreeitemPrototype(
						res, 
						Components.results[res], 
						"result", 
						0, 
						false, 
						true
				));
			}
			
			if(_originalTreeItems.results.length == 0) {
				_originalTreeItems.results = _treeItems.results;
			}
		}
		
		// Load Component Manager entries
		if(aView == "compman") {
			let entries = kCM_SRV.enumerateCategories();
			
			while(entries.hasMoreElements()) {
				_treeItems.compman.push(new XV_TreeitemPrototype(
						entries.getNext().QueryInterface(Components.interfaces.nsISupportsCString), 
						"", 
						"compman_category", 
						0, 
						true, 
						false
				));
			}
			
			if(_originalTreeItems.compman.length == 0) {
				_originalTreeItems.compman = _treeItems.compman;
			}
		}
	}; // end loadTreeitems(aView)
	
	
	/** 
	 * Performs given command. 
	 * 
	 * @param aCmd {String}
	 * @returns {void}
	 */
	this.doCommand = function(aCmd) 
	{
		switch(aCmd) {
			// Quit XPCOMViewer
			case "cmd_quit":
				window.close();
				break;
			
			// Show about dialog
			case "cmd_showAbout":
				let aboutDlg = window.openDialog(
						"chrome://xpcomviewer/content/about.xul", 
						"xpcomviewer-dialog", 
						"chrome,dialog,modal,centerscreen"
				);
				if(aboutDlg) aboutDlg.focus();
				break;
			
			// Show preferences dialog
			case "cmd_showPrefs":
				let prefsDlg = window.openDialog(
						"chrome://xpcomviewer/content/preferences/default.xul", 
						"xpcomviewer-dialog", 
						"chrome,dialog,modal,centerscreen,all"
				);
				if(prefsDlg) prefsDlg.focus();
				break;
			
			// Show specified treeview
			case "cmd_showClasses":
			case "cmd_showIfaces":
			case "cmd_showResults":
			case "cmd_showCompman":
				let type = aCmd.replace("cmd_show", "").toLowerCase();
				this.showView(type);
				break;
			
			// Updates current tree view according to entered filter.
			case "cmd_updateFilter":
				this.cmd_updateFilter();
				break;
			
			// Save report from displayed tree items.
			case "cmd_saveAsHtml":
			case "cmd_saveAsTxt":
			case "cmd_saveAsXml":
				let reportType = aCmd.replace("cmd_saveAs", "").toLowerCase();
				this.cmd_saveAs(reportType);
				break;
			
			// Copy name of selected tree item into clipboard.
			case "cmd_copySel":
				this.cmd_copySel();
				break;
			
			// Add bookmark to selected tree item (if no another is attached).
			case "cmd_bookmarkAdd":
				this.cmd_bookmarkAdd();
				break;
			
			// Remove bookmark of selected tree item (if has one attached)
			case "cmd_bookmarkRemove":
				this.cmd_bookmarkRemove();
				break;
			
			// Generates source code.
			case "cmd_generateCode":
			case "cmd_generateCppCode":
			case "cmd_generateJsCode":
			case "cmd_generatePyCode":
				let codeType = "";
				if(aCmd == "cmd_generateCode") 
					codeType = kPREF_SRV.getCharPref("codegenLang");
				else
					codeType = aCmd.replace("cmd_generate", "").replace("Code", "").toLowerCase();
				
				this.cmd_generateCode(codeType);
				break;
			
			// Show code templates edit dialog
			case "cmd_editCodeTemplates":
				let editCodeDlg = window.openDialog(
						"chrome://xpcomviewer/content/edit-templates-dlg.xul", 
						"xpcomviewer-dialog", 
						"chrome,dialog,centerscreen"
				);
				if(editCodeDlg) editCodeDlg.focus();
				break;
			
			// Search for selected item on selected server.
			case "cmd_searchBugzilla":
			case "cmd_searchDxr":
			case "cmd_searchMdc":
				let serverType = aCmd.replace("cmd_search", "").toLowerCase();
				this.cmd_searchFor(serverType);
				break;
		}
	}; // end doCommand(aCmd)
	
	
	// Below is implementation for single commands
	
	/**
	 *  Updates current tree view according to entered filter.
	 * 
	 * @returns void
	 */
	this.cmd_updateFilter = function()
	{
		if(_filter == null) {
			_filter = document.getElementById("filter-textbox");
		}
		
		var filter = (new String(_filter.value)).trim().toLowerCase();
		
		if(filter == "") {
			if(_filterString != null) {
				_treeItems[_treeType] = _originalTreeItems[_treeType];
				_filterString = null;
				
				//_tree.treeBoxObject.view.selection.clearSelection();
				_tree.view.selection.clearSelection();
				_tree.view = new XV_TreeviewPrototype();
			}
			return;
		} else if(filter == _filterString) {
			return;
		}
		
		let items = null;
		let filtered = new Array();
		
		if(_filterString != null) {
			if(filter.indexOf(_filterString) == 0) {
				items = _treeItems[_treeType];
			}
		}
		
		if(items == null) items = _originalTreeItems[_treeType];
		
		for(let item in items) {
			if(new String(items[item].name).toLowerCase().indexOf(filter) != -1) {
				filtered.push(items[item]);
			}
		}
		
		if(_tree == null) {
			_tree = document.getElementById("tree");
		}
		
		_treeItems[_treeType] = filtered;
		_filterString = filter;
		
		//_tree.treeBoxObject.view.selection.clearSelection();
		_tree.view.selection.clearSelection();
		_tree.view = new XV_TreeviewPrototype();
	}; // end cmd_updateFilter();
	
	
	/**
	 * Save report from displayed tree items.
	 * 
	 * @param aReportType {string} Type of report ['html','txt','xml'].
	 * @returns {void}
	 */
	this.cmd_saveAs = function(aReportType)
	{
		if(_tree == null) {
			_tree = document.getElementById("tree");
		}
		
		// Prepare output document
		let name = _viewLabels[_treeType];
		let doc = null;
		if(aReportType == "xml") {
			doc = Components.classes["@mozilla.org/xml/xml-document;1"].
					createInstance(Components.interfaces.nsIDOMDocument);
			doc = doc.implementation.createDocument("", "items", null);
		} else if(aReportType == "html") {
			doc = "<html>\n<head>\n<title>" + name + "</title>\n</head>\n" +
					"<body>\n<h1>" + name + "</h1>\n<ol>\n";
		} else {
			doc = name + "\n\n";
		}
		
		// Go through all tree items and generate report
		let col1 = _tree.columns.getPrimaryColumn();
		let col2 = _tree.columns.getColumnAt(2);
		
		for(let i=0; i < _tree.view.rowCount; i++) {
			let itemName = _tree.view.getCellText(i, col1);
			let itemValue = _tree.view.getCellText(i, col2);
			
			switch(aReportType) {
				case "xml":
					let elm = doc.createElement("item", "");
					elm.setAttribute("name", itemName);
					elm.setAttribute("value", itemValue);
					elm.setAttribute("id", i);
					doc.documentElement.appendChild(elm);
					break;

				case "html":
					doc = doc + "<li>" + itemName + ", " + itemValue + "</li>\n";
					break;

				case "txt":
					doc = doc + i + ": \"" + itemName + "\", \"" + itemValue + "\"\n";
					break;
			}
		}
		
		// Finish output document
		if(aReportType == "html") {
			doc = doc + "</ol>\n</body>\n</html>";
		}
		
		// Pick target file and write document into it
		let strBundle = document.getElementById("stringbundle");
		let title = strBundle.getString("saveAsDlgLabel");
		
		try {
			let file = xpcomviewer.global.pickFile(title);
			
			if(aReportType == "xml") 
				xpcomviewer.global.writeXmlFile(file, doc);
			else 
				xpcomviewer.global.writeTextFile(file, doc);
		} catch(e) {
			xpcomviewer.global.showAlert("saveAsFailedMsg", "showSaveAsFailedAlert");
		}
	}; // end cmd_saveAs(aReportType)
	
	
	/**
	 *  Copy name of selected tree item into clipboard.
	 * 
	 * @returns {void}
	 */
	this.cmd_copySel = function()
	{
		let itemName = this.getSelectedTreeItemName();
		if(itemName == "") {
			xpcomviewer.global.showAlert("noItemSelectedMsg", "showNoItemSelectedAlert");
			return;
		}
		
		xpcomviewer.global.copyToClipboard(itemName);
	}; // end cmd_copySel()
	
	
	/**
	 * Add bookmark to selected tree item (if no another is attached).
	 * 
	 * @returns {void}
	 */
	this.cmd_bookmarkAdd = function()
	{
		let itemName = this.getSelectedTreeItemName();
		if(itemName == "") {
			xpcomviewer.global.showAlert("noItemSelectedMsg", "showNoItemSelectedAlert");
			return;
		}
		
		if(xpcomviewer.bookmarks.hasBookmark(itemName)) {
			xpcomviewer.global.showAlert("bookmarkAlreadyAttachedMsg", "showBookmarkAlreadyAttachedAlert");
			return;
		}
		
		xpcomviewer.bookmarks.createBookmark(itemName);
	}; // end cmd_bookmarkAdd()
	
	
	/**
	 * Remove bookmark of selected tree item (if has one attached).
	 * 
	 * @returns {void}
	 */
	this.cmd_bookmarkRemove = function()
	{
		let itemName = this.getSelectedTreeItemName();
		if(itemName == "") {
			xpcomviewer.global.showAlert("noItemSelectedMsg", "showNoItemSelectedAlert");
			return;
		}
		
		if(!xpcomviewer.bookmarks.hasBookmark(itemName)) {
			xpcomviewer.global.showAlert("bookmarkNotAttachedMsg", "showBookmarkNotAttachedAlert");
			return;
		}
		
		xpcomviewer.bookmarks.deleteBookmark(itemName);
	}; // end cmd_bookmarkRemove()
	
	
	/**
	 * Generates code.
	 * 
	 * @param aCodeType {string} Specify code: ['cpp','java','js','py'].
	 * @returns {void}
	 */
	this.cmd_generateCode = function(aCodeType)
	{
		// Command 'cmd_generateCode' is not allowed for the 'compman' type of view.
		if(_treeType == "compman") {
			xpcomviewer.global.showAlert("noCodeGenForViewMsg", "showNoCodeGenForViewAlert");
			return;
		}
		
		let item = this.getSelectedTreeItem();
		if(item == null) {
			xpcomviewer.global.showAlert("noItemSelectedMsg", "showNoItemSelectedAlert");
			return;
		}
		
		let tpl = xpcomviewer.templates.getTemplateSource(item.type, aCodeType);
		if(tpl == null || tpl == "") {
			Components.utils.reportError(
					"Template for [" + aCodeType + "][" + item.type + "] not found or empty!"
			);
			// If template is null or empty returns item.name
			//xpcomviewer.global.copyToClibpoard(item.name);
			xpcomviewer.global.showAlert("noTemplateMsg", "showNoTemplateAlert");
			return;
		}
		
		if(item.type == "iface_init" || item.type == "iface_const") {
			// var iface = Components.classes["?1"].getInstance(Components.interfaces.?2);
			// Components.interfaces.?1.?2
			
			let parent = this.getSelectedTreeItemParent();
			if(parent == null) {
				//alert("Parent expected but not found!");
				return;
			}
			
			tpl = tpl.replace("?1", parent.name).replace("?2", item.name);
		}
		else {
			tpl = tpl.replace("?1", item.name);
		}
		
		xpcomviewer.global.copyToClipboard(tpl);
	}; // end cmd_generateCode(aCodeType)
	
	
	/**
	 * Search for selected item on selected server.
	 * 
	 * @param aServerType {string} One of ['bugzilla','dxr','mdc']
	 * @returns {void}
	 */
	this.cmd_searchFor = function(aServerType)
	{
		let itemName = this.getSelectedTreeItemName();
		if(itemName == null) {
			xpcomviewer.global.showAlert("noItemSelectedMsg", "showNoItemSelectedAlert");
			return;
		}
		
		itemName = xpcomviewer.global.encodeUri(itemName);
		
		//let searchUrl = _prefs.getCharPref(aServerType + "Url");
		let searchUrl = "";
		switch(aServerType) {
			case "bugzilla":
				searchUrl = "https://bugzilla.mozilla.org/buglist.cgi?" + 
						"quicksearch=summary:" + itemName + "+content:" + itemName;
				break;
			
			case "dxr":
				searchUrl = "http://dxr.mozilla.org/search.cgi?tree=mozilla-central&" + 
						"string=" + itemName;
				break;
			
			case "mdc":
				searchUrl = "https://developer.mozilla.org/en-US/search?q=" + itemName;
				break;
		}
		
		xpcomviewer.global.openUrl(searchUrl)
	}; // end cmd_searchFor(aServerType)
	
}).apply(xpcomviewer);
